//--------//--------//--------//--------//--------//--------//--------//--------//--------
//
//
// glDevice and tools. 2013
//
//
//--------//--------//--------//--------//--------//--------//--------//--------//--------
#ifndef _COMMON_H_
#define _COMMON_H_

//--------//--------//--------//--------//--------//--------//--------//--------//--------
// tekitou
//--------//--------//--------//--------//--------//--------//--------//--------//--------
#pragma warning(disable:4244)
#pragma warning(disable:4305)
#pragma warning(disable:4005)
#pragma warning(disable:4996)

//--------//--------//--------//--------//--------//--------//--------//--------//--------
// include
//--------//--------//--------//--------//--------//--------//--------//--------//--------
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#include <vector>
#include <stdio.h>
#include <math.h>
#include <ctype.h>

//--------//--------//--------//--------//--------//--------//--------//--------//--------
// lib
//--------//--------//--------//--------//--------//--------//--------//--------//--------
#pragma comment(lib,     "winmm.lib")
#pragma comment(lib,     "user32.lib")
#pragma comment(lib,     "gdi32.lib")
#pragma comment(lib,     "opengl32.lib")
#pragma comment(lib,     "glu32.lib")
#pragma comment(lib,     "dinput8.lib")

//--------//--------//--------//--------//--------//--------//--------//--------//--------
// define
//--------//--------//--------//--------//--------//--------//--------//--------//--------
#define E_RETURN(str, x) { HRESULT hRet = (x); if(hRet != S_OK) { printf("E_RETURN(%08X) : %s\n", hRet, str); return hRet; } }
#define RELEASE(x)       { if(x) (x)->Release(); x = NULL; }
#define ASIZE(x)         ( sizeof(x) / sizeof(x[0]) )

//--------//--------//--------//--------//--------//--------//--------//--------//--------
//
//
// NAMESPACE common START
//
//
//--------//--------//--------//--------//--------//--------//--------//--------//--------
namespace common {
  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  // File
  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  struct File {
    int Read(const char *name, void *buf, int size) {
      int ret = -1;
      FILE *fp = fopen(name, "rb");
      if(fp) {
        ret = fread(buf, 1, size, fp);
        fclose(fp);
      }
      return ret;
    }
  };

  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  // KeyState
  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  struct KeyState {
    int V, H, b0, b1;
    void Update(int in_p = 0xBFFF, int in_n = 0x3FFF) {
      #define KeyState(x) (GetAsyncKeyState(x)  & 0x8000)
      JOYINFO JoyInfo = {0};
      KeyState temp = {
        KeyState(VK_UP)   ? -1 : KeyState(VK_DOWN)  ?  1 : 0,
        KeyState(VK_LEFT) ? -1 : KeyState(VK_RIGHT) ?  1 : 0,
        KeyState('Z'), KeyState('X'),
      };
      if(joyGetPos(JOYSTICKID1, &JoyInfo) == JOYERR_NOERROR) {
        if(!temp.H)  temp.H  = -(JoyInfo.wXpos < in_n) + (JoyInfo.wXpos > in_p);
        if(!temp.V)  temp.V  = -(JoyInfo.wYpos < in_n) + (JoyInfo.wYpos > in_p);
        if(!temp.b0) temp.b0 = (JoyInfo.wButtons & JOY_BUTTON1) > 0;
        if(!temp.b1) temp.b1 = (JoyInfo.wButtons & JOY_BUTTON2) > 0;
      }
      *this = temp;
    }
  };

  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  // vec
  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  typedef float real;
  struct vec {
    real x, y, z;
    vec(real a = 0, real b = 0, real c = 0) : x(a), y(b), z(c) {}
    inline real dot(vec &a) { return x * a.x + y * a.y + z * a.z; }
    inline real len()       { return sqrt(dot(*this)); }
    inline vec  nor()       { real k = 1.0 / len(); return vec(x * k, y * k, z * k); }
    inline void disp()      { printf("%f %f %f : len = %f, \n", x, y, z, len()); }
    #define _vop_(op)         inline vec operator op (const vec &a) { return vec(x op a.x, y op a.y, z op a.z); }
    _vop_(+) _vop_(-) _vop_(*) _vop_(/) _vop_(+=) _vop_(-=) _vop_(*=) _vop_(/=) _vop_(<) _vop_(>) 
  };

  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  // random
  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  struct random {
    long a, b, c;
    random() { a = 1, b = 234567, c = 890123; }
    long  get()  { a += b; b += c; c += a; return (a >> 16); }
    float fget() { return float(get()) / float(0x7FFF); }
  };

  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  //fps : http://www.t-pot.com/program/13_fps/index.html
  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  inline void show_fps() {
    static DWORD    last = timeGetTime();
    static DWORD    frames = 0;
    static char     buf[256] = "";
    DWORD           current;
    current = timeGetTime();
    frames++;
    if(1000 <= current - last) {
      float dt = (float)(current - last) / 1000.0f;
      float fps = (float)frames / dt;
      last = current;
      frames = 0;
      sprintf(buf, "%.02f fps            ", fps);
      printf("%s\n", buf);
    }
  }

  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  //
  // glDevice
  //
  //--------//--------//--------//--------//--------//--------//--------//--------//--------
  struct glDevice {
    enum {
      COLOR_MAX   = 256,
      CIRCLE_DIV  = 32,
      FONT_MAX    = 256,
    };
    
    PIXELFORMATDESCRIPTOR pfd;
    GLYPHMETRICSFLOAT     gmfline[FONT_MAX];
    GLYPHMETRICSFLOAT     gmfpoly[FONT_MAX];
    GLuint                fontline;
    GLuint                fontpoly;
    int                   PixelFormat;
    HWND                  hWnd;
    HDC                   hDC;
    HGLRC                 hRC;
    DWORD                 Width;
    DWORD                 Height;
    GLUquadric           *shape;
    GLfloat               color[COLOR_MAX][4];
    GLfloat              *GetColor(DWORD index) { return color[index % COLOR_MAX]; }
    
    HRESULT Init(HWND handle, DWORD x, DWORD y, BOOL Window = TRUE) {
      ZeroMemory(this, sizeof(glDevice));
      hWnd           = handle;
      Width          = x;
      Height         = y;
      pfd.nSize      = sizeof(PIXELFORMATDESCRIPTOR);
      pfd.nVersion   = 1;
      pfd.dwFlags    = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
      pfd.iPixelType = PFD_TYPE_RGBA;
      pfd.cColorBits = 24;
      pfd.cDepthBits = 16;
      pfd.iLayerType = PFD_MAIN_PLANE;

      SetPixelFormat(hDC, PixelFormat = ChoosePixelFormat( hDC = GetDC(hWnd), &pfd), &pfd);
      hRC = wglCreateContext(hDC);
      wglMakeCurrent(hDC, hRC);
      
      if(!Window) {
        DEVMODE dm;
        ZeroMemory(&dm, sizeof(dm));
        dm.dmSize       = sizeof(DEVMODE);
        dm.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT;
        dm.dmPelsWidth  = Width;
        dm.dmPelsHeight = Height;
        ChangeDisplaySettings(&dm, CDS_FULLSCREEN);
      }
      
      BOOL (WINAPI *wglSwapIntervalEXT)(int) = NULL;
      if( strstr( (char*)glGetString( GL_EXTENSIONS ), "WGL_EXT_swap_control") ) {
        wglSwapIntervalEXT = (BOOL (WINAPI*)(int))wglGetProcAddress("wglSwapIntervalEXT");
        if( wglSwapIntervalEXT ) wglSwapIntervalEXT(1);
      }
      shape = gluNewQuadric();
      
      float  col = 0;
      float dcol = 1.0 / float(COLOR_MAX);
      for(int i = 0 ; i < COLOR_MAX; i++) {
        CreateColor(i, col, 1 - col, col, 0.5);
        col += dcol;
      }
      
      SystemParametersInfo(SPI_SETFONTSMOOTHING, TRUE, NULL, 0 );
      HFONT font = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
      SelectObject(hDC, font);
      fontline = glGenLists(FONT_MAX);
      fontpoly = glGenLists(FONT_MAX);
      wglUseFontOutlines(hDC, 0, FONT_MAX - 1, fontline, 0.0f, 0.1f, WGL_FONT_LINES,   gmfline);
      wglUseFontOutlines(hDC, 0, FONT_MAX - 1, fontpoly, 0.0f, 0.1f, WGL_FONT_POLYGONS, gmfpoly);
      return S_OK;
    }

    void Term() {
      if(fontpoly) glDeleteLists(fontpoly, FONT_MAX);
      if(fontline) glDeleteLists(fontline, FONT_MAX);
      if(hDC)      wglMakeCurrent(hDC, 0);
      if(hRC)      wglDeleteContext(hRC);
      if(hDC)      ReleaseDC(hWnd, hDC);
      if(hWnd)     DestroyWindow(hWnd);
      ZeroMemory(this, sizeof(glDevice));
    }
    
    HRESULT CreateColor(DWORD i, float r, float g, float b, float a) {
      color[i][0] = r; color[i][1] = g; color[i][2] = b; color[i][3] = a;
      return S_OK;
    }
    
    void ShiftColor(DWORD i) {
      GLfloat col = color[i][0]; color[i][0] = color[i][1]; color[i][1] = color[i][2]; color[i][2] = col;
    }
    
    void EnableList(GLenum *w)  { while(*w) glEnable(*w++); }
    void DisableList(GLenum *w) { while(*w) glDisable(*w++); }

    //--------//--------//--------//--------//--------//--------//--------//--------//--------
    // render and draw
    //--------//--------//--------//--------//--------//--------//--------//--------//--------
    void Render(void (CallBack)(glDevice *)) {
      GLenum dpe[] = { GL_DEPTH_TEST, GL_CULL_FACE, GL_NORMALIZE, 0  };
      GLenum epe[] = { GL_BLEND, GL_POLYGON_SMOOTH, 0  };
      DisableList(dpe);
      EnableList(epe);
      glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
      glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
      //glBlendFunc(GL_ONE, GL_SRC_COLOR);
      //glBlendFunc(GL_ONE, GL_DEST_COLOR);
      //glBlendFunc(GL_ZERO, GL_SRC_COLOR);
      //glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ZERO);
      //glBlendFunc(GL_ONE, GL_ONE);
      //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE);
      
      glViewport(0, 0, Width, Height);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0, Width, 0, Height, -1, 1);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glColor3f(1.0, 1.0, 1.0);
      CallBack(this);
      glFinish();
      SwapBuffers(hDC);
    }
    
    void Clear(float r = 0, float g = 0, float b = 0, float a = 0) {
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glClearColor(r, g, b, a);
    }
    
    void SetTransform(float x, float y, float w, float h, float angle) {
      glTranslatef(x, y, 0.0f );
      glRotatef(angle, 0.0f, 0.0f, 1.0f );
      glScalef(w, h, 0);
    }
    
    void DrawText(char *text, INT len, float x, float y, float w = 32, float h = 32, float angle = 0, DWORD color = 200, GLint type = 0) {
      if(!text) return;
      if(len < 0) len = strlen(text);
      glPushMatrix();
        glColor4fv(GetColor(color));
        if(type) glListBase(fontline);
        else     glListBase(fontpoly);
        SetTransform(x, y, w, h, angle);
        glCallLists(len, GL_UNSIGNED_BYTE, text);
      glPopMatrix();
    }

    void DrawDisk(float x, float y, float w, float h, float angle = 0, GLint div = 4, GLint type = 0, DWORD color = 128, float k = 1.0) {
      glPushMatrix();
        SetTransform(x, y, w, h, angle);
        gluQuadricDrawStyle(shape, (type == 0) ? GLU_FILL : (type == 1) ? GLU_SILHOUETTE : GLU_LINE);
        glColor4fv(GetColor(color));
        gluDisk(shape, 0, 1, div, 1);
      glPopMatrix();
    }

    void DrawGrid(float sidew, float ox = 0, float oy = 0, DWORD color = 0, float lw = 0.5) {
      if(sidew <= 0) return;
      glPushMatrix();
        glColor4fv(GetColor(color));
        glLineWidth(lw);
        float width  = float(Width);
        float height = float(Height);
        glBegin(GL_LINES);
          for (float x = fmod(ox, sidew); x < width;  glVertex2f(float(x), 0.0f), glVertex2f(float(x), height), x += sidew);
          for (float y = fmod(oy, sidew); y < height; glVertex2f(0.0f, float(y)), glVertex2f(width,  float(y)), y += sidew);
        glEnd(); 
        glLineWidth(1.0);
      glPopMatrix();
    }
  };


  //---------//---------//---------//---------//---------//---------//---------//---------
  // Make Window
  //---------//---------//---------//---------//---------//---------//---------//---------
  LRESULT WINAPI SimpleWindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
  inline HWND InitWindow(LPCSTR AppName, DWORD w, DWORD h, BOOL Windowed = TRUE, LRESULT WINAPI WindowProc(HWND , UINT , WPARAM , LPARAM ) = SimpleWindowProc) {
    RECT  rc;
    HWND  hWnd     = NULL;
    DWORD Style   =  Windowed ? (WS_OVERLAPPEDWINDOW & ~WS_MAXIMIZEBOX) : WS_POPUP;
    DWORD StyleEx = WS_EX_APPWINDOW | (Windowed ? WS_EX_WINDOWEDGE : 0);
    SetRect( &rc, 0, 0, w, h );
    AdjustWindowRectEx( &rc, Style, FALSE, StyleEx);
    rc.right  -= rc.left;
    rc.bottom -= rc.top;
    static WNDCLASSEX wcex = {
      sizeof(WNDCLASSEX), CS_CLASSDC | CS_VREDRAW | CS_HREDRAW,
      WindowProc, 0L, 0L, GetModuleHandle(NULL), LoadIcon(NULL, IDI_APPLICATION),
      LoadCursor(NULL, IDC_ARROW), (HBRUSH)GetStockObject(BLACK_BRUSH),
      NULL, AppName, NULL
    };
    if(!RegisterClassEx(&wcex)) return NULL;
    DWORD wx = 0;
    DWORD wy = 0;
    if(Windowed) {
      wx = (GetSystemMetrics( SM_CXSCREEN ) - rc.right) / 2 ;
      wy = (GetSystemMetrics( SM_CYSCREEN ) - rc.bottom) / 2;
    }
    hWnd = CreateWindowEx(
      StyleEx, AppName, AppName, Style, wx, wy, 
      rc.right, rc.bottom, NULL, NULL, wcex.hInstance, NULL);
    if(!hWnd) {
      printf("ERROR : Can't Create Window");
      return NULL;
    }

    ShowWindow(hWnd, SW_SHOW);
    UpdateWindow(hWnd);
    Windowed ? (void)0 : ShowCursor(FALSE);
    return hWnd;
  }

  //---------//---------//---------//---------//---------//---------//---------//---------
  // Process Message 
  //---------//---------//---------//---------//---------//---------//---------//---------
  inline BOOL ProcMsg() {
    MSG msg;
    while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
      if(msg.message == WM_QUIT) return FALSE;
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
    return TRUE;
  }

  inline LRESULT WINAPI SimpleWindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    int temp = wParam & 0xFFF0;
    switch(msg) {
    case WM_SYSCOMMAND:
      if(temp == SC_MONITORPOWER || temp == SC_SCREENSAVE) {
        return 0;
      }
      break;
    case WM_IME_SETCONTEXT:
      lParam &= ~ISC_SHOWUIALL;
      break;
    case WM_CLOSE:
    case WM_DESTROY:
      PostQuitMessage(0);
      return 0;
      break;
    case WM_KEYDOWN:
      if(wParam == VK_ESCAPE) PostQuitMessage(0);
      break;
    }
    return DefWindowProc(hWnd, msg, wParam, lParam);
  }

} //namespace common
//--------//--------//--------//--------//--------//--------//--------//--------//--------
//
//
// NAMESPACE common END
//
//
//--------//--------//--------//--------//--------//--------//--------//--------//--------

#endif //_COMMON_H_
